@@ -1,3 +1,14 @@
+#
+# WING - Web-IMAP/NNTP Gateway
+#
+# Wing/Abook.pm
+#
+# Author: Malcolm Beattie, mbeattie@sable.ox.ac.uk
+#
+# This program may be distributed under the GNU General Public License (GPL)
+#
+# 23 Feb 1999  Release version 0.5
+#
 package Wing::Connection;
 use Wing::Shared;
 use SQL;
@@ -39,7 +50,7 @@ sub init_abook_ids {
     # at the end of the list. It will only be included at that point
     # if no earlier entry mentions it explicitly.
     #
-    foreach my $p (maild_decode(split(' ', $list)), "+$DEFAULT_ABOOK") {
+    foreach my $p (maild_decode(split(/:/, $list)), "+$DEFAULT_ABOOK") {
 	my ($active, $abook) = unpack("aa*", $p);
 	my $flags = ($active eq "+") ? ABOOK_ACTIVE : 0;
 	$flags |= ABOOK_OWNED unless $abook =~ m(/); # no slash means we own it
@@ -118,52 +129,121 @@ sub _find_abook_ix ($$) {
     return -1;
 }
 
-#
-# _lookup_alias looks up a list of (potential) aliases in the default
-# address book list (and, if not found there, in the username list).
-# It returns a corresponding list where each successfully found alias
-# is replaced by the email address in the form "First Last <address>"
-# and each unfound string is returned without change.
-#
-
-sub _lookup_alias {
-    my ($conn, @addresses) = @_;
-    my $s = $conn->{maild};
-    my $abook_info = _load_abook_info($s);
-    my @ids = map { $_->[0] } grep { $_->[1] & ABOOK_ACTIVE } @$abook_info;
-    my $id_clause = "id in (" . join(",", @ids) . ")";
-    my $dbh = DBI->connect(@WING_DBI_CONNECT_ARGS);
-    foreach my $alias (@addresses) {
-	next unless is_legal_alias($alias);
-	my $alias_quoted = $dbh->quote($alias);
-	my ($first_name, $last_name, $email) = $dbh->selectrow_array(<<"EOT");
+sub _lookup_one_alias {
+    my ($dbh, $id_clause, $alias) = @_;
+    return $alias unless is_legal_alias($alias);
+    my $alias_quoted = $dbh->quote($alias);
+    my ($first_name, $last_name, $email) = $dbh->selectrow_array(<<"EOT");
 select first_name, last_name, email
 from abook_aliases
 where $id_clause
 and alias = $alias_quoted
 EOT
-	if ($email) {
-	    $alias = "$first_name $last_name <$email>";
-	} elsif ($alias =~ /^[a-z][a-z0-9]{0,7}$/) {
-	    #
-	    # See if it's a username
-	    #
-	    my ($sender) = $dbh->selectrow_array(
-		"select sender from users where username = $alias_quoted"
-	    );
-	    if ($sender) {
-		$alias = $sender;
-	    }
+    if ($email) {
+	if ($email =~ /,/) {
+	    # It's a list
+	    return $email;
+	} else {
+	    return "$first_name $last_name <$email>";
 	}
     }
-    $dbh->disconnect;
-    return @addresses;
+    if ($alias =~ /^[a-z][a-z0-9]{0,7}$/) {
+	#
+	# See if it's a username
+	#
+	my ($sender) = $dbh->selectrow_array(
+	    "select sender from users where username = $alias_quoted"
+	);
+	if ($sender) {
+	    return $sender;
+	}
+    }
+    return undef;
 }
 
-sub _format_address ($$$) {
-    my ($first_name, $last_name, $email) = @_;
+sub _split_addresses {
+    my $addresses = shift;
+    #
+    # XXX For now, we don't worry about commas in comments nor
+    # about trying to allow white space to separate addresses.
+    # We just split on /\s*,\s*/ to find multiple addresses.
+    #
+    return split(/\s*,\s*/, $addresses);
+}
+
+#
+# _lookup_alias looks up a list of (potential) aliases in the default
+# address book list (and, if not found there, in the username list).
+# The list is passed as a single string containing an RFC822 list of
+# addresses (i.e. comma separated).
+# It returns a corresponding list (as a string of the same form) where
+# each successfully found alias is replaced by the email address in the
+# form "First Last <address>" (or just "address1, address2, ..." if the
+# alias resolves to a list--i.e. resolves to a string that contains a
+# comma). Each unfound address is included without change.
+#
+
+sub _lookup_alias {
+    my ($conn, $addresses) = @_;
+    my $s = $conn->{maild};
+    my $abook_info = _load_abook_info($s);
+    my @ids = map { $_->[0] } grep { $_->[1] & ABOOK_ACTIVE } @$abook_info;
+    if (!@ids) {
+	#
+	# We *have* to catch this case of no ids since SQL barfs with
+	# a syntax error on "...where id in ()". Of course, it's an
+	# easy optimisation too.
+	#
+	return $addresses;
+    }
+    my $id_clause = "id in (" . join(",", @ids) . ")";
+
+    my @results;
+    my @in = _split_addresses($addresses);
+    #
+    # Verify that at least one is a possible alias before diving into
+    # connecting to the database and wading through them.
+    #
+    my $may_contain_alias = 0;
+    foreach my $alias (@in) {
+	$may_contain_alias = 1 if is_legal_alias($alias);
+    }
+    return $addresses unless $may_contain_alias;
+
+    #
+    # OK, at least one address in there may be an alias. Connect to
+    # the database and start resolving them.
+    #
+    my @nextin;
+    my $dbh = DBI->connect(@WING_DBI_CONNECT_ARGS);
+
+    # Allow alias recursion up to a given maximum level
+    my $recurse_max = 5;
+
+    for (my $i = 0; @in && $i < $recurse_max; $i++) {
+	while (defined(my $alias = shift @in)) {
+	    my $r = _lookup_one_alias($dbh, $id_clause, $alias);
+	    if (defined($r)) {
+		push(@nextin, _split_addresses($r));
+	    } else {
+		push(@results, $alias);
+	    }
+	}
+	@in = @nextin;
+	@nextin = ();
+    }
+
+    $dbh->disconnect;
+    return join(", ", @results, @in);
+}
+
+sub _resolve_address ($$$$) {
+    my ($conn, $first_name, $last_name, $address) = @_;
     
-    if ($first_name ne "" && $last_name ne "") {
+    my $email = _lookup_alias($conn, $address);
+    if ($email =~ /,/ || $email ne $address) {
+	return $email;
+    } elsif ($first_name ne "" && $last_name ne "") {
 	return "$first_name $last_name <$email>";
     } elsif ($last_name ne "") {
 	return "$last_name <$email>";
@@ -199,8 +279,7 @@ EOT
 	    $comment_enc, $email_enc) = url_encode(@$res);
 	my ($abook_html, $alias_html, $first_name_html, $last_name_html,
 	    $comment_html, $email_html) = map { escape_html($_) } @$res;
-	my $addr_canon = canon_encode(_format_address($first_name, $last_name,
-						      $email));
+	my $addr = join("/", canon_encode($first_name, $last_name, $email));
 	$r->print("<tr>\n");
 	$r->print(<<"EOT") if $inc_abook;
 <td><a href="$url_prefix/abook_search?abook=$abook_enc&key0=alias&op0=matches&val0=*&search=Search">$abook</a></td>
@@ -211,9 +290,9 @@ EOT
 <td>$last_name_html</td>
 <td>$comment_html</td>
 <td>$email_html</td>
-<td><a href="$url_prefix/add_address/$callback/To/$addr_canon">T</a></td>
-<td><a href="$url_prefix/add_address/$callback/Cc/$addr_canon">C</a></td>
-<td><a href="$url_prefix/add_address/$callback/Bcc/$addr_canon">B</a></td>
+<td><a href="$url_prefix/add_address/$callback/To/$addr">T</a></td>
+<td><a href="$url_prefix/add_address/$callback/Cc/$addr">C</a></td>
+<td><a href="$url_prefix/add_address/$callback/Bcc/$addr">B</a></td>
 </tr>
 EOT
     }
@@ -327,7 +406,7 @@ sub cmd_abook_search {
 	my $sql = sprintf("select id,alias,first_name,last_name,comment,email "
 			 ."from abook_aliases where %s and (%s)",
 			  $id_clause, join(" $combiner ", @where));
-	$r->warn("abook_search: $sql"); # debug
+#	$r->warn("abook_search: $sql"); # debug
 
 	#
 	# Prepare and execute the SQL query and pull back the results
@@ -337,7 +416,7 @@ sub cmd_abook_search {
 	$sth->execute
 	     or return wing_error($r, "DBI execute failed: $DBI::errstr");
 	while (my @row = $sth->fetchrow_array) {
-	    $r->warn("row: ", join(", ", @row)); # debug
+#	    $r->warn("row: ", join(", ", @row)); # debug
 	    $row[0] = $id2name{$row[0]};
 	    push(@results, \@row);
 	}
@@ -357,9 +436,11 @@ sub cmd_abook_search {
   <img src="/wing-icons/address-books.gif" border=0 alt="Address Books"></a>
 </td>
 <td><a href="$url_prefix/compose">
-  <img src="/wing-icons/compose.gif" border=0 alt="Compose"></a></td>
+  <img src="/wing-icons/compose.gif" border=0 alt="Compose"></a>
+</td>
 <td><a href="$url_prefix/logout//abook_search">
-  <img src="/wing-icons/logout.gif" border=0 alt="Logout"></a></td>
+  <img src="/wing-icons/logout.gif" border=0 alt="Logout"></a>
+</td>
 <tr></table>
 $info_msg
 <h2 align="center">Address book search</h2>
@@ -451,9 +532,11 @@ sub cmd_abook_open {
   <img src="/wing-icons/address-search.gif" border=0 alt="Address search"></a>
 </td>
 <td><a href="$url_prefix/compose">
-  <img src="/wing-icons/compose.gif" border=0 alt="Compose"></a></td>
+  <img src="/wing-icons/compose.gif" border=0 alt="Compose"></a>
+</td>
 <td><a href="$url_prefix/logout//abook_open">
-  <img src="/wing-icons/logout.gif" border=0 alt="Logout"></a></td>
+  <img src="/wing-icons/logout.gif" border=0 alt="Logout"></a>
+</td>
 <tr></table>
 $info_msg
 <h2 align="center">Address book `$abook_html'</h2>
@@ -463,8 +546,15 @@ EOT
 
     $r->print(<<"EOT");
 <hr>
+<p>To include an entry as an addressee on the Compose screen,
+click "T", "C," or "B" (for "To", "Cc" or "Bcc" respectively).</p>
+<p>To display an entry in full, click on the relevant alias name.
+You can modify, delete or copy an alias in this way.</p>
+<p>To add a new entry, use this button:
 <a href="$url_prefix/abook_entry?abook=$abook_enc">
-  <img src="/wing-icons/add-new-entry.gif" border=0 alt="Add new entry"></a>
+  <img src="/wing-icons/add-new-entry.gif" border=0
+    align="absmiddle" alt="Add new entry"></a>
+</p>
 </body></html>
 EOT
 }
@@ -485,8 +575,8 @@ sub cmd_abook_entry {
 	my $id = _find_abook_id($abook_info, $abook, ABOOK_OWNED)
 	    or return wing_error($r, "invalid address book '$abook'");
 	sql_connect(@WING_DBI_CONNECT_ARGS);
-	local($SIG{__WARN__}) = sub { $r->warn("sql_debug: @_") };
-	sql_debug(1);
+#	local($SIG{__WARN__}) = sub { $r->warn("sql_debug: @_") };
+#	sql_debug(1);
 	sql_table("abook_aliases");
 	sql_select(["count(*)" => \my $count], id => $id, alias => $alias);
 	sql_fetch;
@@ -498,14 +588,16 @@ sub cmd_abook_entry {
 	    # entry already present: update it
 	    sql_update(@fields, id => $id, alias => $alias);
 	    $info_msg .= "\nUpdated entry for alias ".escape_html($alias);
+	    $alias = $first_name = $last_name = $comment = $email = "";
 	} else {
 	    # entry not present: insert it if it's a reasonable alias
 	    if (is_legal_alias($alias)) {
 		sql_insert([id => \$id], [alias => \$alias], @fields);
 		$info_msg .= "\nAdded entry for alias ".escape_html($alias);
+		$alias = $first_name = $last_name = $comment = $email = "";
 	    } else {
-		$info_msg .= "\nCan't add badly formatted alias "
-			     . escape_html($alias);
+		$info_msg .= sprintf("\nIllegal alias: %s\n<br>\n%s\n",
+				     escape_html($alias), $LEGAL_ALIAS_RULES);
 	    }
 	}
 	sql_disconnect;
@@ -513,18 +605,22 @@ sub cmd_abook_entry {
 	sql_connect(@WING_DBI_CONNECT_ARGS);
 	my $rows = sql_delete(from => "abook_aliases", alias => $alias);
 	sql_disconnect;
-	return wing_error($r, "No entry in address book '$abook' for $alias")
-	    unless $rows == 1;
+	if ($rows == 1) {
+	    $info_msg .= "\n<br>Alias $alias has been deleted";
+	    $alias = $first_name = $last_name = $comment = $email = "";
+	} else {
+	    $info_msg .= "\n<br>No entry in address book '$abook' for $alias";
+	}
     } elsif (defined($q{add_to_To})) {
-	my $address = _format_address($first_name, $last_name, $email);
+	my $address = _resolve_address($conn, $first_name, $last_name, $email);
 	printf $s "add_address To %s\n", canon_encode($address);
 	$info_msg .= "\n<br>Added to To header: ".escape_html($address);
     } elsif (defined($q{add_to_Cc})) {
-	my $address = _format_address($first_name, $last_name, $email);
+	my $address = _resolve_address($conn, $first_name, $last_name, $email);
 	printf $s "add_address Cc %s\n", canon_encode($address);
 	$info_msg .= "\n<br>Added to Cc header: ".escape_html($address);
     } elsif (defined($q{add_to_Bcc})) {
-	my $address = _format_address($first_name, $last_name, $email);
+	my $address = _resolve_address($conn, $first_name, $last_name, $email);
 	printf $s "add_address Bcc %s\n", canon_encode($address);
 	$info_msg .= "\n<br>Added to Bcc header: ".escape_html($address);
     }
@@ -549,9 +645,11 @@ sub cmd_abook_entry {
   <img src="/wing-icons/address-search.gif" border=0 alt="Address Search"></a>
 </td>
 <td><a href="$url_prefix/compose">
-  <img src="/wing-icons/compose.gif" border=0 alt="Compose"></a></td>
+  <img src="/wing-icons/compose.gif" border=0 alt="Compose"></a>
+</td>
 <td><a href="$url_prefix/logout//abook_entry">
-  <img src="/wing-icons/logout.gif" border=0 alt="Logout"></a></td>
+  <img src="/wing-icons/logout.gif" border=0 alt="Logout"></a>
+</td>
 </tr></table>
 $info_msg
 <h1 align="center">Entry for address book
@@ -611,7 +709,9 @@ sub cmd_abook_list {
 	    if ($key eq "create_add") {
 		my $abook = $q{abook};
 		#
-		# Sanity check address book name
+		# Sanity check address book name. Don't use
+		# is_legal_abook_name yet since we want to allow
+		# names of the form "username/foo".
 		#
 		if (length($abook) > 256) {
 		    return wing_error($r, "address book name too long: $abook");
@@ -632,10 +732,10 @@ sub cmd_abook_list {
 			    # XXX Ought to allow explicit "ourusername/foo"
 			    $info_msg .= "\n<br><strong>No such address book "
 					. "and can't create it</strong>";
-			} elsif ($abook eq "") {
-			    $info_msg .= "\n<br><strong>Can't create address book with empty name</strong>";
-			} elsif ($abook =~ /[^\w!$%^*()+=:;@~#?.,-]/) {
-			    $info_msg .= "\n<br><strong>Bad character in address book name: not created</strong>";
+			} elsif (!is_legal_abook_name($abook)) {
+			    $info_msg .= sprintf(
+	"\n<br><strong>Illegal address book name: %s</strong><br>\n%s\n",
+	 escape_html($abook), $LEGAL_ABOOK_RULES);
 			} else {
 			    # Doesn't exist: create it
 			    sql_select(["nextval('abook_ids_seq')" => \$id]);
@@ -668,7 +768,8 @@ sub cmd_abook_list {
 		}
 	    } elsif ($key eq "search") {
 		return redirect($r, "$url_prefix/abook_search?"
-				."search=Search&alias=".url_encode($q{alias}));
+				  . "search=Search&key0=alias&op0=is&val0="
+				  . url_encode($q{alias}));
 	    } elsif ($key eq "update_save") {
 		$do_save = 1;
 	    }
@@ -686,7 +787,7 @@ sub cmd_abook_list {
 	    print $s "username\n";
 	    chomp(my $username = <$s>);
 	    my $dbh = DBI->connect(@WING_DBI_CONNECT_ARGS);
-	    my $list_q = $dbh->quote(join(" ", @list));
+	    my $list_q = $dbh->quote(join(":", @list));
 	    my $done = $dbh->do(
 	      "update options set abooklist=$list_q where username='$username'"
 	    );
@@ -708,14 +809,21 @@ sub cmd_abook_list {
 <body>
 <table><tr>
 <td><a href="$url_prefix/$abook_return">
-  <img src="/icons/back.gif" border=0 alt="Back"></a></td>
+  <img src="/icons/back.gif" border=0 alt="Back"></a>
+</td>
 <td><a href="$url_prefix/help/abook_list">
-  <img src="/wing-icons/help.gif" border=0 alt="Help"></a></td>
+  <img src="/wing-icons/help.gif" border=0 alt="Help"></a>
+</td>
 <td><a href="$url_prefix/abook_search">
   <img src="/wing-icons/address-search.gif" border=0 alt="Address Search"></a>
 </td>
+<td><a href="$url_prefix/abook_import">
+  <img src="/wing-icons/import-address-book.gif" border=0
+    alt="Import Address Book"></a>
+</td>
 <td><a href="$url_prefix/logout//abook_list">
-  <img src="/wing-icons/logout.gif" border=0 alt="Logout"></a></td>
+  <img src="/wing-icons/logout.gif" border=0 alt="Logout"></a>
+</td>
 </tr></table>
 $info_msg
 <h2 align="center">Address books</h2>
@@ -725,7 +833,6 @@ EOT
 	my ($id, $flags, $name) = @{$abook_info->[$ix]};
 	my $name_html = escape_html($name);
 	my $name_enc = url_encode($name);
-	my $checked; # XXX delete when commented-out html removed
 	my $active_html;
 	my $up_html;
 	my $down_html;
@@ -746,7 +853,10 @@ EOT
 	if ($flags & ABOOK_OWNED) {
 	    $buttons_html = <<"EOT";
 <a href="$url_prefix/abook_perms/$ix">
-  <img src="/wing-icons/permissions.gif" border=0 alt="Permissions"></a></td>
+  <img src="/wing-icons/permissions.gif" border=0 alt="Permissions"></a>
+EOT
+	    $buttons_html .= <<"EOT" unless $name eq "personal";
+</td>
 <td><a href="$url_prefix/abook_adjust/delete/$ix">
   <img src="/wing-icons/delete.gif" border=0 alt="Delete"></a>
 EOT
@@ -823,8 +933,22 @@ sub cmd_abook_adjust {
 	print $s "abook_drop $ix\n";
 	maild_set($s, "message", "Address book dropped from search list");
     } elsif ($cmd eq "delete" && $ix >= 0 && $ix <= $max) {
-	# XXX Delete address book from abook_ids and abook_perms if empty
-	maild_set($s, "message", "Address book deletion not yet implemented");
+	my ($id, $flags, $name) = @{$abook_info->[$ix]};
+	if (!($flags & ABOOK_OWNED)) {
+	    maild_set($s, "message",
+		      "You cannot delete an address book which you do not own");
+	} elsif ($name eq "personal") {
+	    maild_set($s, "message",
+		      "You cannot delete your personal address book");
+	} else {
+	    my $dbh = DBI->connect(@WING_DBI_CONNECT_ARGS);
+	    $dbh->do("delete from abook_aliases where id = $id");
+	    $dbh->do("delete from abook_perms where id = $id");
+	    $dbh->do("delete from abook_ids where id = $id");
+	    $dbh->disconnect;
+	    print $s "abook_drop $ix\n";
+	    maild_set($s, "message", "Address book $name has been deleted");
+	}
     } elsif ($cmd eq "activate" && $ix >= 0 && $ix <= $max) {
 	my $flags = $abook_info->[$ix]->[1] | ABOOK_ACTIVE;
 	print $s "abook_flags $ix $flags\n";
@@ -856,6 +980,7 @@ sub cmd_abook_perms {
     }
     my $abook_html = escape_html($abook);
     my $abook_enc = url_encode($abook);
+    my $group = maild_get($s, "group");
 
     my %q = $r->args;
     my %ok_user;
@@ -957,18 +1082,20 @@ To drop this permission, use this button:
   <img align="absmiddle" src="/wing-icons/drop-permission.gif" border=0 alt="Drop Permission"></a>
 EOT
     } else {
-	$r->print("The following usernames have been granted access:\n",
+	$r->print("<h4>Usernames that have been granted access</h4>\n",
 		  join(", ", @users_html),
-		  "\n<hr>\n",
-		  "The following groups have been granted access:\n",
+		  "\n",
+		  "<h4>Groups that have been granted access</h4>\n",
 		  join(", ", @groups_html),
 		  "\n<hr>\n");
 	$r->print(<<"EOT");
-To drop permission for a particular user or group, click on the name.
-<br>
+<h4>Changing permissions</h4>
+<p>You are in group $group</p>
+<p>To drop permission for a particular user or group, click on
+its name in the appropriate list above.</p>
+<p>To grant permission to a particular user or group, enter the name
+below and press the "Allow" button.</p>
 <form>
-To grant permission to a particular user or group, enter the name
-below and press the "Allow" button.
 <br>
 Allow access to
 <select name="type" size=1>
@@ -988,11 +1115,13 @@ EOT
 }
 
 sub cmd_add_address {
-    my ($conn, $callback, $hdr, $address) = @_;
+    my ($conn, $callback, $hdr, $first_name, $last_name, $email) = @_;
+    ($first_name, $last_name, $email) =
+	canon_decode($first_name, $last_name, $email);
     my $r = $conn->{request};
     my $s = $conn->{maild};
 
-    $address = canon_decode($address);
+    my $address = _resolve_address($conn, $first_name, $last_name, $email);
     $callback = canon_decode($callback);
     #
     # Sanity check
@@ -1008,4 +1137,205 @@ sub cmd_add_address {
     return redirect($r, "$conn->{url_prefix}/$callback");
 }
 
+#
+# For importing address books we need MIME::Parser
+#
+use MIME::Parser;
+
+sub _receive_abook_upload {
+    my $r = shift;
+    my ($type, $abook, $lines);
+    my $size = $r->header_in("Content-Length");
+    if ($size > $ABOOK_IMPORT_SIZE_LIMIT) {
+	return <<"EOT";
+Address book uploaded was too large. Your client wanted to send
+$size bytes (including MIME overhead data) but the size limit is
+$ABOOK_IMPORT_SIZE_LIMIT bytes.
+EOT
+    }
+    my $buffer = "";  # must reset $buffer or Apache::read appends to it
+    my $didread = $r->read($buffer, $size);
+    if ($didread != $size) {
+	return "client failed to upload correct data size";
+    }
+    
+#    $r->warn("raw data: $buffer");#debug
+    my $parser = MIME::Parser->new(output_to_core => "ALL");
+    $buffer =~ /^--(.*)\r\n/
+	or return "Badly formatted MIME form from client";
+    $buffer = "Content-Type: multipart/form-data;boundary=$1\r\n\r\n$buffer";
+    my $ent = $parser->parse_data($buffer);
+    $buffer = "";
+    
+    my $num = $ent->parts;
+    $r->warn("$num MIME parts");#debug
+    for (my $i = 0; $i < $num; $i++) {
+	my $part = $ent->parts($i);
+	my $name = $part->head->mime_attr("content-disposition.name");
+	$r->warn("part $i name=$name");#debug
+	if ($name eq "abook") {
+	    $abook = $part->body->[0];
+	    $r->warn("abook=$abook");#debug
+	    if (!is_legal_abook_name($abook)) {
+		return "Illegal address book name chosen. $LEGAL_ABOOK_RULES";
+	    }
+	} elsif ($name eq "type") {
+	    $type = $part->body->[0];
+	    $r->warn("type=$type");#debug
+	    return "Illegal format chosen" unless $type eq "pine";
+	} elsif ($name eq "file") {
+	    $lines = $part->body;
+	    $r->warn("lines=$lines");#debug
+	}
+    }
+
+    if (!defined($lines)) {
+	return "No address book file uploaded";
+    }
+
+    return ($abook, $type, $lines);
+}
+
+sub _next_pine_abook_entry {
+    my $lines = shift;
+    my $line;
+    do {
+	$line = shift @$lines;
+    } while $line =~ /^#/;
+    return () unless $line;
+    $line .= shift @$lines while $lines->[0] =~ s/^\s+//;
+    $line =~ tr/\r\n//d; # either, not just both together
+    my ($alias, $full_name, $email, $fcc, $comment) = split(/\t/, $line);
+    next if $alias eq "";
+    $email .= "";	# ensure it's not undef
+    $comment .= "";	# ensure it's not undef
+    for ($email) {
+	s/^\(//;
+	s/\)$//;
+    }
+    my $first_name = "";
+    my $last_name = "";
+    if ($full_name =~ /^"(.*)"$/) {
+	$last_name = $1;
+    } elsif ($full_name =~ /^(.*?),\s*(.*)$/) {
+	$first_name = $2;
+	$last_name = $1;
+    } else {
+	$last_name = $full_name;
+    }
+    return ($alias, $first_name, $last_name, $comment, $email);
+}
+
+sub cmd_abook_import {
+    my $conn = shift;
+    my $r = $conn->{request};
+    my $s = $conn->{maild};
+    my $url_prefix = $conn->{url_prefix};
+    my $info_msg = info_message_html($s);
+    my ($type, $abook, $lines);
+
+    if ($r->method eq "POST") {
+	($abook, $type, $lines) = _receive_abook_upload($r);
+	if (!defined($lines)) {
+	    $info_msg = $abook;
+	    $abook = "";
+	    goto show_form;
+	}
+
+	my $abook_info = _load_abook_info($s);
+	my $id = _find_abook_id($abook_info, $abook, ABOOK_OWNED);
+	if ($id) {
+	    $info_msg .= <<"EOT";
+Address book $abook already exists. Please give the name of an
+address book you want to create to hold the newly imported data.
+EOT
+	    goto show_form;
+	}
+
+	sql_connect(@WING_DBI_CONNECT_ARGS);
+#	local($SIG{__WARN__}) = sub { $r->warn("sql_debug: @_") };#debug
+#	sql_debug(1);#debug
+
+	# Create the new address book
+	print $s "username\n";
+	chomp(my $username = <$s>);
+
+	sql_select(["nextval('abook_ids_seq')" => \$id]);
+	sql_fetch or $r->log_error("abook_ids_seq failed");
+	sql_insert(into => "abook_ids",
+		   [id => \$id], [username => \$username], [tag => \$abook]);
+	printf $s "abook_add %s %s %s\n",
+	    maild_encode($id, ABOOK_OWNED|ABOOK_ACTIVE, $abook);
+	# $abook_info now out of date (would need to _load_abook_info if nec)
+
+	sql_table("abook_aliases");
+	my ($alias, $first_name, $last_name, $comment, $email);
+	my @fields = ([id => \$id],
+		      [alias => \$alias],
+		      [first_name => \$first_name],
+		      [last_name => \$last_name],
+		      [comment => \$comment],
+		      [email => \$email]);
+
+	my $count = 0;
+	while (($alias, $first_name, $last_name, $comment, $email) =
+		_next_pine_abook_entry($lines))
+	{
+	    if (!is_legal_alias($alias)) {
+		$info_msg .= "\nOmitted entry with illegal alias: '$alias'. ";
+	    } else {
+		$count++;
+		sql_insert(@fields);
+		$r->warn("a=$alias fn=$first_name ln=$last_name c=$comment e=$email");#debug
+	    }
+	}
+	sql_disconnect;
+	$info_msg .= "\nImported $count entries into new address book $abook";
+	$abook = "";
+    }
+
+  show_form:
+
+    $info_msg = "<strong>$info_msg</strong>";
+    my $abook_html = escape_html($abook);
+    dont_cache($r, "text/html");
+    $r->print(<<"EOT");
+<html><head><title>Import address book</title>
+<body>
+<table><tr>
+<td><a href="$url_prefix/list">
+  <img src="/icons/back.gif" border=0 alt="Back"></a>
+</td>
+<td><a href="$url_prefix/abook_list">
+  <img src="/wing-icons/address-books.gif" border=0 alt="Address Books"></a>
+</td>
+<td><a href="$url_prefix/logout//abook_import">
+  <img src="/wing-icons/logout.gif" border=0 alt="Logout"></a>
+</td>
+</tr></table>
+$info_msg
+<h1 align="center">Import address book</h1>
+<form method="POST" action="$url_prefix/abook_import"
+  enctype="multipart/form-data">
+<table>
+<tr>
+  <td>Import
+    <select size=1 name="type">
+      <option value="pine" selected>Pine</option>
+    </select>
+    format address book file</td>
+  <td><input type="file" name="file"></td>
+</tr>
+<tr>
+  <td>and create new address book with name</td>
+  <td><input name="abook" value="$abook_html"></td>
+</tr>
+</table>
+<br>
+<input type="submit" value="Import">
+</form>
+</body></html>
+EOT
+}
+
 1;
