@@ -31,3 +31,13 @@ create table abook_aliases (
 	comment		text	not null,
 	email		text	not null
 );
+
+--
+-- Let root create new address books (so that an initial personal
+-- address book can be created when each account is created) and
+-- let root read the aliases and perms tables so it can back them up.
+--
+grant all on abook_ids, abook_ids_seq, abook_ids_idx to httpd;
+grant all on abook_ids, abook_ids_seq, abook_ids_idx to root;
+grant all on abook_perms, abook_aliases to httpd;
+grant all on abook_perms, abook_aliases to root;
